
// Example: launch of Saturn V simulation
// Model source: Speckhart: A Guide to CSMP. Prentice-Hall, 1976, [p38]
// (Not valid: results differ, needs corrections)

// Uses simple (not optimized) implementation of Runge-Kutta 4
// Needs linking with math library (GCC option -lm)
// Author: Petr Peringer, FIT TU Brno, 2019

#include <stdio.h>
#include <math.h>

// System: three stage rocket
#define SIZE 2                  // state vector size
const double mass[3] = { 148820, 33205, 8137 }; // slug
const double flow[3] = { 930, 81.49, 14.75 };   // slug/s
const double vout[3] = { 8060, 13805, 15250 };  // foot/s
const double drag[3] = { 510, 460, 360 };       // drag coef. * stage crossection [foot^2]
const double burnt[3] = { 150, 359, };  // burn time [s]
const double radearth = 3960.0*5280.0;  // Earth radius [foot]
double mph;
double acc;
double miles;
inline int stage(double time) {
    if(time < burnt[0]) return 0;
    if(time < (burnt[0]+burnt[1])) return 1;
    return 2;
}
inline double sqr(double x) { return x*x; }
// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    double m;
    int st = stage(t);
    switch(st) {
        case 0: m = mass[0]+mass[1]+mass[2] - flow[0]*t;
                break;
        case 1: m = mass[1]+mass[2] - flow[1]*(t-burnt[0]);
                break;
        case 2: m = mass[2] - flow[2]*(t-burnt[0]-burnt[1]);
                break;
    }
    double g = 32.17*sqr(radearth/(radearth+y[0]));
    double ro = 0.00238*exp(-y[0]/24000);       // air density [slug/foot^2]
    acc = (flow[st]*vout[st]-m*g-0.5*drag[st]*ro*sqr(y[1]))/m;
    mph = y[1] * 60.0/88.0;
    miles = y[0]/5280.0;
    yin[0] = y[1];
    yin[1] = acc;
}

void RK4step(double t, double h, double *y);

// Experiment control
int main()
{
    double y[SIZE] = { 0, 0 };
    double stepsize = 4;

    double t = 0;
    while (t < 988) {
        printf("%8.3f % -11g % -11g  % g\n", t, y[0]/5280.0, y[1], acc);
        RK4step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
    }
    printf("%8.3f % -11g % -11g  % g\n", t, y[0]/5280.0, y[1], acc);
    return 0;
}

//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}

